#include "module.h"

static void ReadMonitor(Config *c,const char *key,int keylen,LSMONITOR *res)
{
	ASSERT(READABLE(c,sizeof(*c)));
	ASSERT(WRITEABLE(c,sizeof(*c)));
	ASSERT(READABLE(key,keylen+1));
	ASSERT(WRITEABLE(res,sizeof(*res)));
	ASSERT(READABLE(res,sizeof(*res)));

	ConfigState cs(c,key,keylen);
	if (GetRCString(c->key,c->line,"",MAX_LINE_LENGTH))
		*res = LSGetMonitor(c->line);
}

static void ReadCoord(Config *c,const char *key,int keylen,LONG *res,LONG max)
{
	ASSERT(READABLE(c,sizeof(*c)));
	ASSERT(WRITEABLE(c,sizeof(*c)));
	ASSERT(READABLE(key,keylen+1));
	ASSERT(WRITEABLE(res,sizeof(*res)));
	ASSERT(READABLE(res,sizeof(*res)));

  ConfigState cs(c,key,keylen);
	*res = GetRCCoordinate(c->key,*res,max);
}

static void ReadBool(Config *c,const char *key,int keylen,bool *res)
{
	ASSERT(READABLE(c,sizeof(*c)));
	ASSERT(WRITEABLE(c,sizeof(*c)));
	ASSERT(READABLE(key,keylen+1));
	ASSERT(WRITEABLE(res,sizeof(*res)));
	ASSERT(READABLE(res,sizeof(*res)));

  ConfigState cs(c,key,keylen);
	*res = FALSE!=GetRCBoolDef(c->key,*res?TRUE:FALSE);
}

static void ReadInt(Config *c,const char *key,int keylen,int *res)
{
	ASSERT(READABLE(c,sizeof(*c)));
	ASSERT(WRITEABLE(c,sizeof(*c)));
	ASSERT(READABLE(key,keylen+1));
	ASSERT(WRITEABLE(res,sizeof(*res)));
	ASSERT(READABLE(res,sizeof(*res)));

  ConfigState cs(c,key,keylen);
	*res = GetRCInt(c->key,*res);
}

#ifndef NO_SCRIPT_ACTIONS
static void ReadAction(Config *c,const char *key,int keylen,ScriptAction *res)
{
	ASSERT(READABLE(c,sizeof(*c)));
	ASSERT(WRITEABLE(c,sizeof(*c)));
	ASSERT(READABLE(key,keylen+1));
	ASSERT(WRITEABLE(res,sizeof(*res)));
	ASSERT(READABLE(res,sizeof(*res)));

  ConfigState cs(c,key,keylen);
	if (GetRCLine(c->key,c->line,MAX_LINE_LENGTH,""))
  {
    if (*res) delete[] *res;
    *res = new char[lstrlen(c->line)+1];
    lstrcpy(*res,c->line);
  }
}

static void ReadActions(Config *c,const char *key,int keylen,ScriptActionSet *res,int nActions)
{
	ASSERT(READABLE(c,sizeof(*c)));
	ASSERT(WRITEABLE(c,sizeof(*c)));
	ASSERT(READABLE(key,keylen+1));
	ASSERT(WRITEABLE(res,sizeof(*res)));
	ASSERT(READABLE(res,sizeof(*res)));

  ConfigState cs(c,key,keylen);

  const char **pName = ScriptActionStrings;
  int nSet = 0, iAction;

  if (*res==NULL)
  {
    *res = new ScriptAction[nActions];
    ZeroMemory(*res,nActions*sizeof(ScriptAction));
  }

  for (iAction=0; iAction<nActions; ++iAction,++pName)
  {
    ReadAction(c,*pName,lstrlen(*pName),(*res)+iAction);
    if ((*res)[iAction]!=NULL) ++nSet;
  }

  if (nSet==0)
  {
    delete[] *res;
    *res = NULL; /* used as flag */ 
  }
}

static void FreeActions(ScriptActionSet *res,int nActions)
{
  if (*res!=NULL)
  {
    while (nActions--)
    {
      if (res[nActions])
        delete[] (*res)[nActions];
    }

    delete *res;
  }
}
#endif /* NO_SCRIPT_ACTIONS */ 

#ifndef NO_SKIN
static void ReadBitmap(Config *c,const char *key,int keylen,HBITMAP *res,SIZE *sz)
{
	ASSERT(READABLE(c,sizeof(*c)));
	ASSERT(WRITEABLE(c,sizeof(*c)));
	ASSERT(READABLE(key,keylen+1));
	ASSERT(WRITEABLE(res,sizeof(*res)));
	ASSERT(READABLE(res,sizeof(*res)));
	ASSERT(WRITEABLE(sz,sizeof(*sz)));
	ASSERT(READABLE(sz,sizeof(*sz)));

  ConfigState cs(c,key,keylen);
  int w,h;

	if (GetRCString(c->key,c->line,"",MAX_LINE_LENGTH))
  {
	  *res = LoadLSImage(c->line,NULL);
	  GetLSBitmapSize(*res,&w,&h);
	  sz->cx = w;
	  sz->cy = h;
  }
}

static void ReadRect(Config *c,const char *key,int keylen,RECT *res,SIZE *max)
{
	ASSERT(READABLE(c,sizeof(*c)));
	ASSERT(WRITEABLE(c,sizeof(*c)));
	ASSERT(READABLE(key,keylen+1));
	ASSERT(WRITEABLE(res,sizeof(*res)));
	ASSERT(READABLE(res,sizeof(*res)));

  ConfigState cs(c,key,keylen);
  int args;

	if (GetRCLine(c->key,c->line,MAX_LINE_LENGTH,""))
  {
	  args = CommandTokenize(c->line,c->tokens,4,0);
	  if (args == 1) {
		  res->left     =
		  res->top      =
		  res->right    =
		  res->bottom   = atoi(c->one);
	  } else {
		  res->left     = atoi(c->one);
		  res->top      = atoi(c->two);
		  res->right    = atoi(c->three);
		  res->bottom   = atoi(c->four);
	  }
  }

  ReadCoord(c,"Left",  4,&res->left,  max->cx);
  ReadCoord(c,"Top",   3,&res->top,   max->cy);
  ReadCoord(c,"Right", 5,&res->right, max->cx);
  ReadCoord(c,"Bottom",6,&res->bottom,max->cy);
}

static void ReadSkin(Config *c,const char *key,int keylen,lsEdgeMap *res)
{
	ASSERT(READABLE(c,sizeof(*c)));
	ASSERT(WRITEABLE(c,sizeof(*c)));
	ASSERT(READABLE(key,keylen+1));
	ASSERT(WRITEABLE(res,sizeof(*res)));
	ASSERT(READABLE(res,sizeof(*res)));

  ConfigState cs(c,key,keylen);
  bool bTile = res->Tile!=FALSE;

  ReadBitmap(c,"Bitmap",6,&res->Bitmap,&res->Size);
  ReadRect(c,"Edge",4,&res->Edge,&res->Size);
  ReadBool(c,"Tile",4,&bTile);
  res->Tile = bTile?TRUE:FALSE;
}
#endif

//
// Module-wide settings
//

void ReadConfigMod(ModInfo *m,Config *c,const char *key,int keylen)
{
	ASSERT(READABLE(c,sizeof(*c)));
	ASSERT(WRITEABLE(c,sizeof(*c)));
	ASSERT(READABLE(key,keylen+1));
	ASSERT(WRITEABLE(m,sizeof(*m)));
	ASSERT(READABLE(m,sizeof(*m)));

  ConfigState cs(c,key,keylen);

#ifdef SINGLE_WINDOW
	m->bWaitForBox = false;
	ReadBool(c,"InBox",5,&m->bWaitForBox);
#endif

#ifndef NO_SCRIPT_ACTIONS
  ReadActions(c,"On",2,&m->Actions,ScriptActionModuleCount);

  { /* use system value, or 1/4 sec for win95 */ 
    UINT t = 1000/4;
    #define SPI_GETMOUSEHOVERTIME      0x0066
    SystemParametersInfo(SPI_GETMOUSEHOVERTIME,0,&t,0);
    m->HoverTime = t;
  }
  ReadInt(c,"HoverTime",9,&m->HoverTime);
#endif
}

void FreeConfigMod(ModInfo *m)
{
	ASSERT(WRITEABLE(m,sizeof(*m)));
	ASSERT(READABLE(m,sizeof(*m)));

#ifndef NO_SCRIPT_ACTIONS
  FreeActions(&m->Actions,ScriptActionModuleCount);
#endif
}

//
// Window-specific settings
//

void ReadConfigWnd(WndInfo *w,Config *c,const char *key,int keylen)
{
	ASSERT(READABLE(c,sizeof(*c)));
	ASSERT(WRITEABLE(c,sizeof(*c)));
	ASSERT(READABLE(key,keylen+1));
	ASSERT(WRITEABLE(w,sizeof(*w)));
	ASSERT(READABLE(w,sizeof(*w)));

  ConfigState cs(c,key,keylen);

	if (w->hParentWnd != Module.hBottomParentWnd)
	{
		w->Monitor = LSMONITOR_NONE;
		GetWindowRect(w->hParentWnd,&w->rMonitor);
	}
	else
	{
		w->Monitor = LSGetMonitor(NULL);
		ReadMonitor(c,"OnMonitor",9,&w->Monitor);
		LSGetMonitorScreenRect(w->Monitor,&w->rMonitor);
	}

	w->Pos.x = w->Pos.y = 0;
	ReadCoord(c,"X",1,&w->Pos.x,w->rMonitor.right-w->rMonitor.left);
	ReadCoord(c,"Y",1,&w->Pos.y,w->rMonitor.bottom-w->rMonitor.top);

	if (w->hParentWnd != Module.hBottomParentWnd)
	{
		w->Size.cx = w->rMonitor.right-w->rMonitor.left-w->Pos.x;
		w->Size.cy = w->rMonitor.bottom-w->rMonitor.top-w->Pos.y;
	}
	else
	{
		w->Size.cx = w->Size.cy = 100;
	}
	ReadCoord(c,"Width",5,&w->Size.cx,w->rMonitor.right-w->rMonitor.left);
	ReadCoord(c,"Height",6,&w->Size.cy,w->rMonitor.bottom-w->rMonitor.top);

	w->Alpha = 255;
	ReadCoord(c,"Alpha",5,&w->Alpha,255);

	w->bTrueAlpha = false;
  /* always use fake alpha if true is not available */ 
  if (dUpdateLayeredWindow!=NULL)
	  ReadBool(c,"TrueAlpha",9,&w->bTrueAlpha);

	w->bOnTop = false;
	ReadBool(c,"OnTop",5,&w->bOnTop);

	w->bHidden = false;
	ReadBool(c,"Hidden",6,&w->bHidden);

#ifndef SINGLE_WINDOW
  w->HoverTime = 0;
  ReadInt(c,"HoverTime",9,&w->HoverTime);

#ifndef NO_SCRIPT_ACTIONS
  w->Actions = NULL;
  ReadActions(c,"On",2,&w->Actions,ScriptActionWindowCount);
#endif

  w->Group = 0;
  ReadInt(c,"Group",5,&w->Group);
#endif /* SINGLE_WINDOW */ 

#ifndef NO_SKIN
  ReadSkin(c,"Skin",0,&w->emSkin);
#endif
}

void FreeConfigWnd(WndInfo *w)
{
	ASSERT(WRITEABLE(w,sizeof(*w)));
	ASSERT(READABLE(w,sizeof(*w)));

#if !defined(NO_SCRIPT_ACTIONS) && !defined(SINGLE_WINDOW)
  FreeActions(&w->Actions,ScriptActionWindowCount);
#endif

#ifndef NO_SKIN
  if (w->emSkin.Bitmap)
    DeleteObject(w->emSkin.Bitmap);
#endif
}

#ifndef SINGLE_WINDOW
int ParseBoxProc(void *elem,const char *line,DWORD conf)
{
	WndInfo *w = (WndInfo *)elem;
	Config *c = (Config *)conf;
	ASSERT(READABLE(w,sizeof(WndInfo)));
	ASSERT(WRITEABLE(w,sizeof(WndInfo)));
	ASSERT(READABLE(c,sizeof(Config)));
	ASSERT(WRITEABLE(c,sizeof(Config)));
	ASSERT(READABLE(line,sizeof(char)));
	char token[MAX_LINE_LENGTH],*ptr;

	if (!GetToken(line,token,&line,FALSE)) return PSL_FAIL|PSL_CONTINUE;

	w->Name = new char[lstrlen(token)+1];
	lstrcpy(w->Name,token);
	w->hParentWnd = Module.hBottomParentWnd;
	ReadConfigWnd(w,c,token,lstrlen(token));

	if (!GetToken(line,token,&line,FALSE)) return PSL_OK|PSL_CONTINUE;

	if (*token>='A')
	{
		w->Monitor = LSGetMonitor(token);
		LSGetMonitorScreenRect(w->Monitor,&w->rMonitor);
		if (!GetToken(line,token,&line,FALSE)) return PSL_OK|PSL_CONTINUE;
	}

	w->Pos.x = ParseCoordinate(token,w->Pos.x,w->rMonitor.right-w->rMonitor.left);
	if (!GetToken(line,token,&line,FALSE)) return PSL_OK|PSL_CONTINUE;

	w->Pos.y = ParseCoordinate(token,w->Pos.y,w->rMonitor.bottom-w->rMonitor.top);
	if (!GetToken(line,token,&line,FALSE)) return PSL_OK|PSL_CONTINUE;

	w->Size.cx = ParseCoordinate(token,w->Size.cx,w->rMonitor.right-w->rMonitor.left);
	if (!GetToken(line,token,&line,FALSE)) return PSL_OK|PSL_CONTINUE;

	w->Size.cy = ParseCoordinate(token,w->Size.cy,w->rMonitor.bottom-w->rMonitor.top);
	if (!GetToken(line,token,&line,FALSE)) return PSL_OK|PSL_CONTINUE;

	if (*token=='#')
	{
    int g = 0;
		ptr = token;
		while (*++ptr)
		{
			if (*ptr>='0' && *ptr<='9')
			{
				g = g*10 + *ptr-'0';
			}
			else switch (*ptr)
			{
			case 't':
			case 'T': w->bOnTop = !w->bOnTop; break;
			case 'h':
			case 'H': w->bHidden = !w->bHidden; break;
      case 'l':
      case 'L': w->bTrueAlpha = !w->bTrueAlpha; break;
			}
		}
    if (g) w->Group = g;
	}

	return PSL_OK|PSL_CONTINUE;
}

void FreeWndProc(void *elem,DWORD)
{
	WndInfo *w=(WndInfo*)elem;
	ASSERT(READABLE(w,sizeof(WndInfo)));
	ASSERT(WRITEABLE(w,sizeof(WndInfo)));
	DestroyWindow(w->hWnd);
  FreeConfigWnd(w);
	if (w->Name) delete[] w->Name;
}
#endif /* SINGLE_WINDOW */ 


